
/** \file

  Wavelet spectral analysis for non-stationary data
  (e.g., equity close prices).  Non-stationary signals are signals
  that do repeat infinitely, as sine and cosine based signals
  do.

  The documentation in this file is formatted for doxygen
  (see www.doxygen.org).

<h4>
   Copyright and Use
</h4>

<p>
   You may use this source code without limitation and without
   fee as long as you include:
</p>
<blockquote>
     This software was written and is copyrighted by Ian Kaplan, Bear
     Products International, www.bearcave.com, 2002.
</blockquote>
<p>
   This software is provided "as is", without any warranty or
   claim as to its usefulness.  Anyone who uses this source code
   uses it at their own risk.  Nor is any support provided by
   Ian Kaplan and Bear Products International.
<p>
   Please send any bug fixes or suggested source changes to:
<pre>
     iank@bearcave.com
</pre>

  @author Ian Kaplan

 */

#include <assert.h>
#include <stdio.h>

#include <vector>

#include "signalUtil.h"
#include "spectrum.h"
#include "yahooTS.h"
#include "line.h"



void calcReturn( double *ret, const double *close, size_t N )
{
  for (size_t i = 1; i <= N; i++)
    ret[i-1] = (close[i] - close[i-1])/close[i-1];
}



/** \function 

  The entry point for code to test the wavelet packet transform.

  The code in main provides a simple example of how to call the
  wavelet packet transform code.

  This test case can be used with various wavelet functions
  (e.g., Haar, linear interpolation, Daubechies D4, polynomial
  interpolation).  In the context of filters, the polynomial
  interpolation wavelets (see poly.h and polyHaar.h) are very
  bad choices.  The polynomial interpolation wavelets do not
  divide the spectrum into high and low frequency components.
  Instead, the power spectrum shows one peak.  I'm not sure
  if this means that the polynomial interpolation wavelets
  are bad for all applications.  For example, another way to
  look at wavelets is as approximation functions.  For
  example, wavelet compression relies approximation functions.
  In this case polynomial wavelets may be a good choice for
  some data sets.

*/
int
main()
{
  const size_t NClose = 513;
  double close[NClose];

  const char *path = "..\\data\\equities\\";
  yahooTS ts( path );
  const char *stock = "amat";

  size_t n = NClose;
  ts.getTS( stock, close, n, yahooTS::Close );

  if (n != NClose) {
    printf("Error reading file %s%s\n", path, stock );
    return 0;
  }

  const size_t N = 512;
  double ret[ N ];

  calcReturn( ret, close, N );

  // prVec( ret, N );

  // linear interpolation wavelet
  line<double *> w;

  double *ptr = ret;  
  w.forwardTrans( ptr, N );

  std::vector<double> powerVec;
  spectrum::spectralCalc( ret, N, powerVec );

  //
  // Print out the power spectrum
  //
  size_t len = powerVec.size();
  for (size_t i = 0; i < len; i++) {
    // printf("%2d, %7.4f\n", i, powerVec[i] );
  }

  double d1[N], d2[N];

  spectrum::copyBands( d1, ret, N, 0, 7 );
  spectrum::copyBands( d2, ret, N, 8, 9 );

  ptr = d1;
  w.inverseTrans( ptr, N );

  ptr = d2;
  w.inverseTrans( ptr, N );

  //
  // Print out the lower half of the spectrum
  //
  // signalUtil::prVec( d1, N );

  //
  // Print out the upper half of the spectrum
  //
  signalUtil::prVec( d2, N );

  return 0;
}
