
#include <math.h>
#include <stdio.h>

#include "signalUtil.h"


/**
  Generate a signal composed of a sum of sine waves. 
  The sine waves summed are in decreasing magnitude and
  decreasing frequency.
 */
void signalUtil::gen_freqMix( double *vecX, double *vecY, size_t N)
{
  const double PI = 3.1415926535897932384626433832795;
  const double range = 2 * PI;
  const double incr = range / (double)N;

  double point = 0.0;
  int i;
  for (i = 0; i < N; i++) {
    vecX[i] = point;
    vecY[i] = 4 * sin( 64 * point ) + 
              2 * sin( 32 * point ) + 
              1 * sin( 16 * point ) +
            0.5 * sin( 8  * point );
    // printf("%7.4g, %7.4g\n", vecX[i], vecY[i] );
    point = point + incr;
  }
} // gen_freqMix



/**
  Generate a signal composed of one or more sine waves.
 */
void signalUtil::gen_sinCombo( double *vecX, double *vecY, size_t N )
{
  const double PI = 3.1415926535897932384626433832795;
  const double range = 1 * PI;
  const double incr = range / (double)N;

  double point = 0.0;

  int stepCnt = 0;
  int i;
  for (i = 0; i < N; i++) {
    vecX[i] = point;
    // vecY[i] = sin( 64 * point ) + sin( 32 * point ) + sin( 16 * point ); 
    // vecY[i] = sin( 16 * PI * point ) + sin( 4 * PI * point ); 
    vecY[i] = sin( 4 * PI * point ); 

    // printf("x[%2d] = %7.4g, y[%2d] = %7.4g\n", i, vecX[i], i, vecY[i] );
    // printf("%7.4g, %7.4g\n", vecX[i], vecY[i] );
    // printf("%d, %7.4g\n", i, vecY[i] );

    point = point + incr;
  }
}

/** 
  Given two points {x<sub>1</sub>, y<sub>1</sub>} and
  {x<sub>2</sub>, y<sub>2</sub>}, where x<sub>1</sub> and
  x<sub>2</sub> are positive integer values, calculate the <i>y</i>
  value for a point {x, y}, where <i>x</i> is a positive integer such
  that x<sub>1</sub> &lt; x &lt; x<sub>2</sub> (e.g., x lies between
  x<sub>1</sub> and x<sub>2</sub>).


  The "two-point equation" for a line given x<sub>1</sub>,
  y<sub>1</sub> and x<sub>2</sub>, y<sub>2</sub> is

     <pre>
     .          y<sub>2</sub> - y<sub>1</sub>
     (y - y<sub>1</sub>) = -------- (x - x<sub>1</sub>)
     .          x<sub>2</sub> - x<sub>1</sub>
     </pre>

     Solving for y

     <pre>
     .    y<sub>2</sub> - y<sub>1</sub>
     y = -------- (x - x<sub>1</sub>) + y<sub>1</sub>
     .    x<sub>2</sub> - x<sub>1</sub>
     </pre>


*/
double signalUtil::new_y( size_t x1, double y1, 
			  size_t x2, double y2,
			  size_t newX )
{
  double newY = 0.0;

  newY = (((y2 - y1)/(x2 - x1)) * (newX - x1)) + y1;
  return newY;
} // new_y


/**
  Generate a sawtooth wave centered around zero on the x-axis.
  There is probably a simpler way to do this, but hey, this works.

  @param vec array to be initialized with the sawtooth wave.
  @param N number of elements in <i>vec</i>
  @param numCycles number of sawtooth wave cycles
  @param amplitude the amplitude on the positive y-axis.  So
         if the amplitude is 1.5, the wave will have a point
         at 1.5 on the y-axis for the positive part of the 
         cycle and at -1.5 for the negative part of the
	 cycle.

 */
void signalUtil::sawToothWave( double *vec,
			       size_t N,
			       size_t numCycles,
			       double amplitude )
{
  size_t spacing = N / numCycles;
  size_t step = spacing / 4;
  bool positive = true;

  size_t x1, x2;
  double y1, y2;

  for (size_t i = 0; i < N; i = i + spacing) {
    x1 = i;
    x2 = i + step;
    for (size_t j = 1; j <= 4; j++) {
      switch (j) {
	case 1: {
	  y1 = 0.0;
	  y2 = amplitude;
	}
	break;
	case 2:{
	  y1 = amplitude;
	  y2 = 0.0;
	}
	break;
	case 3:{
	  y1 = 0.0;
	  y2 = -amplitude;
	}
	break;
	case 4:{
	  y1 = -amplitude;
	  y2 = 0.0;
	}
	break;
      } // switch
      vec[x1] = y1;
      for (size_t k = x1+1; k < x2; k++) {
	vec[k] = new_y( x1, y1, x2, y2, k );
      }
      x1 = x2;
      x2 = x2 + step;
    }
  }
} // sawToothWave



/**
  Add vector A to vector B and place the result in vector C.
  N is the size of the vectors.
 */
void signalUtil::addSignal( double *C, 
			    const double *A,
			    const double *B,
			    const size_t N )
{
  for (size_t i = 0; i < N; i++) {
    C[i] = A[i] + B[i];
  }
}


/**
  
  Print a vector of doubles whose length is len.

 */
void signalUtil::prCoords( double *vecX, double *vecY, size_t len )
{
  for (int i = 0; i < len; i++) {
    printf("%7.4f  %7.4f\n", vecX[i], vecY[i] );
  }
}


/**
  Print a vector of doubles
 */
void signalUtil::prVec( double *vec, size_t len )
{
  for (int i = 0; i < len; i++) {
    printf("%4d  %7.4f\n", i, vec[i] );
  }
} // prVec



/**
  If two vectors are equal, return true, otherwise
  return false.
 */
bool signalUtil::vecEqual( const double *d1,
			   const double *d2,
			   const size_t N )
{
  bool rslt = true;

  for (size_t i = 0; i < N; i++) {
    if (d1[i] != d2[i]) {
      rslt = false;
      break;
    }
  }
  return rslt;
}
