
/** \file

  Wavelet spectral analysis, using stationary signals (e.g., signals
  composed from sine waves that can be viewed as infinitely
  repeating).

  By calculating the energy of each wavelet coefficient band
  a spectral plot, similar to a Fourier spectral plot can be
  generated.  The code in this file tests this for stationary
  combinations of sin waves.

  The documentation in this file is formatted for doxygen
  (see www.doxygen.org).

<h4>
   Copyright and Use
</h4>

<p>
   You may use this source code without limitation and without
   fee as long as you include:
</p>
<blockquote>
     This software was written and is copyrighted by Ian Kaplan, Bear
     Products International, www.bearcave.com, 2002.
</blockquote>
<p>
   This software is provided "as is", without any warranty or
   claim as to its usefulness.  Anyone who uses this source code
   uses it at their own risk.  Nor is any support provided by
   Ian Kaplan and Bear Products International.
<p>
   Please send any bug fixes or suggested source changes to:
<pre>
     iank@bearcave.com
</pre>

  @author Ian Kaplan

 */

#include <assert.h>
#include <stdio.h>

#include <vector>

#include "signalUtil.h"
#include "spectrum.h"

#include "haar.h"
#include "daub.h"
#include "line.h"

#include "poly.h"


void genSawTooth( double *vec, size_t N)
{
  double *A = new double[N];
  double *B = new double[N];

  signalUtil::sawToothWave( A, N, 8, 1.5 );  
  signalUtil::sawToothWave( B, N, 32, 0.25 );
  signalUtil::addSignal( vec, A, B, N );
  delete [] A;
  delete [] B;
}


/** \function 

  The entry point for code to test the wavelet packet transform.

  The code in main provides a simple example of how to call the
  wavelet packet transform code.

  This test case can be used with various wavelet functions
  (e.g., Haar, linear interpolation, Daubechies D4, polynomial
  interpolation).  In the context of filters, the polynomial
  interpolation wavelets (see poly.h and polyHaar.h) are very
  bad choices.  The polynomial interpolation wavelets do not
  divide the spectrum into high and low frequency components.
  Instead, the power spectrum shows one peak.  I'm not sure
  if this means that the polynomial interpolation wavelets
  are bad for all applications.  For example, another way to
  look at wavelets is as approximation functions.  For
  example, wavelet compression relies approximation functions.
  In this case polynomial wavelets may be a good choice for
  some data sets.

*/
int
main()
{
  const size_t N = 1024;
  // const size_t N = sizeof( data ) / sizeof( double );
  double vecX[N], vecY[N];

  // signalUtil::gen_freqMix( vecX, vecY, N );
  signalUtil::gen_sinCombo( vecX, vecY, N );

  // genSawTooth( vecY, N );
  // signalUtil::prVec( vecY, N );

  // signalUtil::prVec( vecY, N );
  // signalUtil::prCoords( vecX, vecY, N );

  // The "Haar" transform
  // haar<double * > w;

  // Daubechies D4 wavelet
  // Daubechies<double * > w;

  // linear interpolation wavelet
  // line<double *> w;

  //
  // Polynomial interpolation wavelet (not recommended)
  //
  poly<double *> w;

  double *ptr = vecY;  
  w.forwardTrans( ptr, N );

  std::vector<double> powerVec;
  spectrum::spectralCalc( vecY, N, powerVec );

  //
  // Print out the power spectrum
  //
  size_t len = powerVec.size();
  for (size_t i = 0; i < len; i++) {
    // printf("%2d, %7.4f\n", i, powerVec[i] );
  }

  double d1[N], d2[N];

  spectrum::copyBands( d1, vecY, N, 0, 6 );
  spectrum::copyBands( d2, vecY, N, 7, 10 );

  ptr = d1;
  w.inverseTrans( ptr, N );

  ptr = d2;
  w.inverseTrans( ptr, N );

  //
  // Print out the lower half of the spectrum
  //
  // signalUtil::prCoords( vecX, d1, N );
  // signalUtil::prVec( d1, N );

  //
  // Print out the upper half of the spectrum
  //
  signalUtil::prCoords( vecX, d2, N );
  // signalUtil::prVec( d2, N );

  return 0;
}
