
#ifndef _WAVELETDEBUG_H_
#define _WAVELETDEBUG_H_

#include <stdio.h>

/**
   This class contains functions that are useful in debugging
   wavelet code.
 */
class waveletDebug {
   public:
   
   /**
      Print the result of a wavelet transform so that each coefficient
      spectrum is printed in its own block, showing the structure of
      the result.
   */
   void pr_ordered( const double *coef, const size_t len)
   {
      const char *format = "%7.4f ";
      printf("{");
      int cnt = 0;
      int num_in_freq = 0;
      for (int i = 0; i < len; i++) {
         printf(format, coef[i]);
         cnt++;
         if (num_in_freq == 0) {
            printf("\n");
            cnt = 0;
            num_in_freq = 1;
         }
         else if (cnt == num_in_freq) {
            printf("\n");
            cnt = 0;
            num_in_freq = num_in_freq * 2;
         }
      }
      printf("}\n");
   } // pr_ordered

   /**
      Print wavelet transform intermediate results.

      The wavelet transform calculates a set of averages and
      coefficients (differences).  These are calculated recursively.
      For example, for a time series with size 32, the first iteration
      of the recursive wavelet transform will calculate a set of 16
      averages and 16 coefficients (differences).

      The next set of differences and avarages will be calculated at
      the start of the array (0..15) and will consist of 8 averages
      and 8 differences.

      This function is passed a value for <tt>N</tt> which is the size
      of the average and difference spectrum.  It prints each of these.

   */
   void pr_intermediate( const double *vec, int N)
   {
      const char* fmt = "%7.4f ";
      int i;
      int half = N >> 1;
      
      printf(" ave = ");
      for (i = 0; i < half; i++) {
         printf(fmt, vec[i] );
      }
      printf("\n");
      printf("coef = ");
      for (i = half; i < N; i++) {
         printf(fmt, vec[i] );
      }
      printf("\n");
   } // pr_intermediate


   /**
      Print the first N elements of <tt>vec</tt>.  If N is less than
      the length of <tt>vec</tt> the print a dividing mark (e.g., "|")
      and print the rest of the array.

      This function is useful for displaying the flow of the
      wavelet computation.
   */
   void pr_vector( const double *vec, const size_t N, const size_t len)
   {
     const char* fmt = "%7.4f ";

     int i;
     for (i = 0; i < N; i++) {
       printf(fmt, vec[i] );
     }

     if (len != N) {
       printf(" | ");
       
       for (i = N; i < len; i++) {
	 printf(fmt, vec[i] );
       }
     }
     printf("\n");
   } // pr_vector

};  // waveletDebug


#endif
